{$A+,B-,D+,E+,F-,G+,I+,L+,N+,P+,Q-,R-,S-,T-,V+,X+,Y+}

(* SURFACES
	 --------
(C) 1999 by Arne von Wyss - arsene@vonwyss.ch

Rev: 19991002

This source code is freeware for non-commercial use and provided AS-IS.
You are not allowed to use or spread incomplete (inclusive this header)
copies of this code. Use at your own risk. If you do not agree with these
statements, delete this source code immediately. Please contact the
author via email for questions or bug reports.

Bugs, Limitations, Debugging
----------------------------

No known bugs. However, the complexity of one surface cannot exceed 8190
rectangles. Such complex surfaces are not expected by me, but if this really
isn't enough you may want to tile several TSurfaces and use them as one.
During tests I almost never got more than 50 rectangles doing normal GUI
usage and operation.

Defining the symbol DEBUG will enable a variable holding the maximum surface
complexity of the current session: "MaxRect" *)
{DEFINE DEBUG}

{$IFDEF DMPI} {only cache hlines when memory is not too much of an issue}
{$DEFINE CACHE}
{$ENDIF}

{The whole ~GUI~ area management, no matter if shapes, mouse click checks or
clipping, are done with ~TSurface~s. The surfaces offer everything to
work with them; one can copy surfaces, invert, clear and combine surfaces,
plus rectangles or borders can be added or removed from a surface.

The surface data structures follow strong rules, thus allowing to make
assumptions when accessing its data. For more information, check the
~TSurface~ help.}

unit Surfaces;

(* Unit interface is HelpGen enabled *)

interface

uses
	Objects,Events;

type
	PSurface=^TSurface;
	{Mode to use for surface combinations using ~SurfaceCombine~.

	 cmOR    : Parts visible in one of both canas will remain visible.
	 cmAND   : Only parts visible in both stay.
	 cmORNOT : All the parts visible in the second canvas will be removed.
	 cmANDNOT: All the parts invisible in the second canvas will be removed.
	 cmXOR   : Only parts visible in one or the other, but not both, are visible.

	Bit masks:
	 cmOR     cmAND    cmORNOT  cmANDNOT cmXOR
	 0011     0011     0011     0011     0011
	 0101     0101     0101     0101     0101
	                     
	 0111     0001     1011     0010     0110}
	TCombineMode=(cmOR,cmAND,cmXOR,cmORNOT,cmANDNOT);
	{HLine model consists of the same elements as a TRect; B.Y is used internally.

	See ~SurfaceGetHLine~ and ~SurfaceResetHLine~ for an example on its usage.}
	THLine=TRect;
	{Array of rectangles as used for ~TSurface.SurfaceRect~.}
	TRectArray=array[0..8189] of TRect;
	{Surfaces are objects that keep a list of rectangles. Surfaces, no matter if
	of the same size or not, can be combined with different modes, plus bars and
	borders can be drawn. The list of rectangles will always comply to the the
	folloging rules, allowing very efficient algorithms when checking them:

	 1) the rectangles are all in the image range (0,0)-(Width-1,Height-1)
	 2) the rectangles are always optimal horizontal stripes
	 3) the rectangle list is sorted top-left to bottom-right (like we read)

	Surfaces are designed to be inherited by other objects. That's why most
	fields and methods have the prefix "Surface". The exceptions are methods
	that are likely to be used in objects derived from a TSurface: the three
	methods to check if a certain coordinate or rectangle is inside or oversects
	the surface, plus the Width and Height fields which you can use for your
	objecs as well (reading, change them by SurfaceResize).

	The surface object is designed to be used as graphics clipping shape. To make
	the speed loss of using complex surfaces smaller the programmer using the
	TSurface or any of its descendants is encouraged to directly access the list
	of rectangles (for reading purposes only). Coordinates checks should work
	like the following pseudo-code example:
	

	function PointInside(X,Y);
		repeat
			while (not EndOfRectList) and (Rect.B.Y<Y) do
				NextRect;
			if EndOfRectList or (Rect.A.Y>Y) then
				return False;
			if (X>=Rect.A.X) and (X<=Rect.B.X) then
				return True;
			NextRect;
		until False;

	Note that, if the surface is empty, the check is most likely to be faster
	than a normal clipping rectangle test, while almost no supplementary code is
	executed if a sigle rectangle is checked! That makes this sort of clipping
	efficient without really slowing down.

	You can see how easy it is to find a point inside the surface: as long as the
	lower border of the recangle checked is above the checked coordinate we take
	the next rectangle. If now the upper border is out of bound we know that no
	more rectangles will fit, so we are sure that the point is not inside the
	surface (rules 2 and 3). If we're still not out, so we have to check if the
	X-coordinate matches. If yes, then the poinmt is inside the rectangle; if not
	we have to start over again using the next rectangle. This can also be easily
	converted to assembler (it does compile and works):
	

							les di,self
							mov cx,es:[di].SurfaceCount
							jcxz @False
							mov ax,X
							mov bx,Y
							les di,es:[di].SurfaceRect
							jmp @StartLoop
	@FindRect:	dec cx
							jz @False
							add di,type TRect
	@StartLoop:	cmp es:[di].TRect.B.Y,bx
							jl @FindRect
	@Check:			cmp es:[di].TRect.A.Y,bx
							jg @False
							cmp es:[di].TRect.A.X,ax
							jg @FindRect
							cmp es:[di].TRect.B.X,ax
							jl @FindRect
	@EndCheck:	mov ax,1
							jmp @End
	@False:			sub ax,ax
	@End:

	If you want to have the best possible efficiency, then you should inherit
	your objects from TSurface. This will make operations on the clipping surface
	easy and transparent, while you have the possibility to quickly access the
	rectangles list.}
	TSurface=object(TEventObject)
	private
		Allocated: Integer;
		{$IFDEF CACHE}
		Cache: Pointer;
		InvalidCache: Integer;
		function GetCache(Y: Integer): Integer;
		procedure InvCache(Y: Integer);
		{$ENDIF}
		procedure Allocate(AAlloc: Integer);
		procedure Add(Pos,X1,Y1,X2,Y2: Integer);
		procedure Remove(Pos: Integer);
		procedure VisibleBar(Start: Integer; var ToCheck: Integer; X1,Y1,X2,Y2: Integer);
		procedure Sort(L,R: Integer);
	public
		{Rectangle list. You may directly read rectangles, but changing the list is
		not adviced. Use ~SurfaceCount~ to stay within the range of available
		rectangles.}
		SurfaceRect: ^TRectArray;
		{Number of TRects in the ~SurfaceRect~ list.}
		SurfaceCount,
		{Width of the surface. Read-Only, use ~SurfaceResize~ to change the size.}
		Width,
		{Height of the surface. Read-Only, use ~SurfaceResize~ to change the size.}
		Height: Integer;
		{Inializes a new surface with the specified size, depending on Visible its
		filled or empty. Other ways to initialize a surface are ~InitCopy~ and
		~InitCopyInvert~.}
		constructor Init(AWidth,AHeight: Integer; Visible: Boolean);
		{Initialize a new surface by making a copy of an existig surface. Other ways
		to initialize a surface are ~Init~ and ~InitCopyInvert~.}
		constructor InitCopy(ASurface: PSurface);
		{Initialize a new surface by making an inverted copy of an existig surface.
		Other ways to initialize a surface are ~Init~ and ~InitCopy~.}
		constructor InitCopyInvert(ASurface: PSurface);
		{Copy a surface over the current data.}
		procedure SurfaceCopy(ASurface: PSurface);
		{Swap the data with another surface. Tis one is very useful for quick surface
		switches as the swap process does not reallocate any memory.}
		procedure SurfaceSwap(ASurface: PSurface);
		{Resize the canvas to a new size. Visible specifies how added space should be
		assigned.}
		procedure SurfaceResize(AWidth,AHeight: Integer; Visible: Boolean);
		{Combine two surfaces. Aside from the combination type given by a ~TCombineMode~
		you can specify wether the border should be added (OR) or removed (AND) it the
		surfaces don't cover each other.}
		procedure SurfaceCombine(ASurface: PSurface; X,Y: Integer; Mode: TCombineMode; Border: Boolean);
		{Draw a bar into the surface. The bar can be visible or hidden. To get real
		drawing operations on a ~TSurface~ obect, use a ~TSurfaceCanvas~.

		The coordinates will always be kept inside the specified size, modifying the
		clipping region in negative or over the border is not possible.

		From the point of view of performance, cmOR and cmANDNOT are the fastest while
		cmXOR is the slowest operation. Thus, you should always try to use only the
		fastest two if possible.}
		procedure SurfaceBar(X1,Y1,X2,Y2: Integer; Visible: Boolean);
		{Only keeps the data inside the rectangle given by (X1,Y1)-(X2,Y2), everything
		outside will be set to visible or invisible, depending on Visible.}
		procedure SurfaceBorder(X1,Y1,X2,Y2: Integer; Visible: Boolean);
		{Invert the content of the surface.}
		procedure SurfaceInvert;
		{Clear the surface contect with the given value.}
		procedure SurfaceClear(Visible: Boolean);
		{Initialize HLine to trace the surface. This will initialize a ~THline~ object
		with the values needed to get the visible segments of a horizontal line.

		See ~SurfaceGetHLine~ for an example on its usage.}
		procedure SurfaceResetHLine(var AHLine: THLine);
		{Get the segment of a horizontal line initialized by ~SurfaceResetHLine~. As
		long as TRUE is returned as result, X1 and X2 will hold the next segment of
		the line.

		Assuming we have a variable L: ~THLine~, the following example will only draw
		lines specified by the segments returned.
		

		L.A.X:=X1;
		L.A.Y:=Y;
		L.B.X:=X2;
		SurfaceResetHLine(L);
		while SurfaceGetHLine(L,X1,X2) do
		Graphics.Line(X1,Y,X2,Y);}
		function SurfaceGetHLine(var AHLine: THLine; var X1,X2: Integer): Boolean;
		{Checks wether two surfaces are equal.}
		function SurfaceEqual(ASurface: PSurface): Boolean;
		{Returns TRUE if the specified point is visible inside the surface. Coordinates
		outside of the surface will be returned as FALSE.}
		function PointIncluded(X,Y: Integer): Boolean;
		{Check wether a certain rectangle is completely visible inside the surface.}
		function RectIncluded(X1,Y1,X2,Y2: Integer): Boolean;
		{Check wether a certain rectangle is partly visible inside the surface.}
		function RectOversect(X1,Y1,X2,Y2: Integer): Boolean;
		{Destroy the surface and free its allocated memory.}
		destructor Done; virtual;
		{The SurfaceRectHLineCut method is used to split rectangles. This will violate
		the least important rule: the rectangles will no longer be optimally collated
		(until a call to SurfaceBar re-collates). However, in certain situations the
		blocks must not "overlap" when moved (for BitBlit for instance), which can be
		achieved using this method. The returned value is the position of the first
		split rectangle if any, else it'll be SurfaceCount. This can be used to "know"
		if the splittiong has to be restarted further back.

		Don't use SurfaceRectHLineCut unless you know what you are doing!}
		function SurfaceRectHLineCut(Y: Integer): Integer;
	end;

{$IFDEF DEBUG}
var
	UseCount,MaxRect: Integer;
	UseTotal: Longint;
{$ENDIF}

{Returns the surface that is covered by two given rectangles. If they do not
touch, the function will retirn FALSE and the values in X5,Y5,X6,Y6 will not
be defined (but they may have been destroyed).}
function RectIntersect(X1,Y1,X2,Y2,X3,Y3,X4,Y4: Integer; var X5,Y5,X6,Y6: Integer): Boolean;

implementation

uses
	GUIUtils;

const
	{$IFDEF DPMI}
	UpSize=62; {Takes about 0.5K per block}
	{$ELSE}    {More agressive memory allocation in DMPI}
	UpSize=8; {Takes about 0.06K per block}
	{$ENDIF}

{$IFDEF CACHE}
var
	Caching: Boolean;
{$ENDIF}

{The different memory allocation strategy makes sure that under REAL mode unused
memory is released quicker; when a TSurface is empty it will release all memory.
In DPMI performance is more an issue because memory allocation under DPMI is
slower and one can be pretty sure that there is more than the 400k of the real
mode available. Note that all memory gets released on TSurface destruction.}

function Intersect(X1,Y1,X2,Y2,X3,Y3,X4,Y4: Integer; var X5,Y5,X6,Y6: Integer): Boolean;
begin {Coordinates must be 1<=2 & 3<=4 in order to work properly!}
	{$IFDEF DEBUG}
	if (X1>X2) or (X3>X4) or (Y1>Y2) or (Y3>Y4) then
		RunError(211);
	{$ENDIF}
	X5:=Maximum(X1,X3);
	Y5:=Maximum(Y1,Y3);
	X6:=Minimum(X2,X4);
	Y6:=Minimum(Y2,Y4);
	Intersect:=(X5<=X6) and (Y5<=Y6);
end;

function RectIntersect(X1,Y1,X2,Y2,X3,Y3,X4,Y4: Integer; var X5,Y5,X6,Y6: Integer): Boolean;
var
	I: Integer;
begin
	if X1>X2 then begin
		I:=X1;
		X1:=X2;
		X2:=I;
	end;
	if Y1>Y2 then begin
		I:=Y1;
		Y1:=Y2;
		Y2:=I;
	end;
	if X3>X4 then begin
		I:=X3;
		X3:=X4;
		X4:=I;
	end;
	if Y3>Y4 then begin
		I:=Y3;
		Y3:=Y4;
		Y4:=I;
	end;
	RectIntersect:=Intersect(X1,Y1,X2,Y2,X3,Y3,X4,Y4,X5,Y5,X6,Y6);
end;

{$IFDEF CACHE}
function TSurface.GetCache(Y: Integer): Integer;
type
	TCache=array[0..32761] of Integer;
	PCache=^TCache;
var
	Nr: Integer;
begin
	{$IFDEF DEBUG}
	if (Height>0) and (not Assigned(Cache)) then
		RunError;
	{$ENDIF}
	if (Y<0) or (Y>=Height) or (SurfaceCount=0) or (not Assigned(Cache)) then begin
		GetCache:=SurfaceCount;
		Exit;
	end;
	if Caching then begin
		GetCache:=0;
		Exit;
	end;
	Caching:=True;
	if InvalidCache<=Y then begin
		FillChar(PCache(Cache)^[InvalidCache],(Height-InvalidCache)*2,$FF);
		InvalidCache:=Height;
		Nr:=-1;
	end else
		Nr:=PCache(Cache)^[Y];
	if Nr<0 then begin
		Nr:=0;
		while Nr<SurfaceCount do begin
			with SurfaceRect^[Nr] do
				if A.Y>Y then begin
					Nr:=SurfaceCount;
					Break;
				end else if B.Y>=Y then
					Break;
			Inc(Nr);
		end;
		PCache(Cache)^[Y]:=Nr;
	end;
	Caching:=False;
	GetCache:=Nr;
end;

procedure TSurface.InvCache(Y: Integer);
begin
	InvalidCache:=Maximum(0,Minimum(InvalidCache,Y));
end;
{$ENDIF}

procedure TSurface.Allocate(AAlloc: Integer);
var
	Temp: Pointer;
begin
	if AAlloc=Allocated then
		Exit;
	{$IFDEF DEBUG}
	if AAlloc<SurfaceCount then
		RunError(201);
	{$ENDIF}
	if AAlloc<>0 then
		GetMem(Temp,AAlloc*SizeOf(TRect))
	else
		Temp:=nil;
	if Assigned(SurfaceRect) then begin
		if Assigned(Temp) then
			MoveL(SurfaceRect^,Temp^,SurfaceCount*SizeOf(TRect) div 4);
		FreeMem(SurfaceRect,Allocated*SizeOf(TRect));
	end;
	SurfaceRect:=Temp;
	Allocated:=AAlloc;
end;

procedure TSurface.Add(Pos,X1,Y1,X2,Y2: Integer);
begin
	{$IFDEF DEBUG}
	if (Pos<0) or (Pos>SurfaceCount) then
		RunError(201);
	{$ENDIF}
	if SurfaceCount=Allocated then
		Allocate(Allocated+UpSize);
	if Pos<SurfaceCount then
		MoveL(SurfaceRect^[Pos],SurfaceRect^[Pos+1],(SurfaceCount-Pos)*SizeOf(TRect) div 4);
	with SurfaceRect^[Pos] do begin
		A.X:=X1;
		A.Y:=Y1;
		B.X:=X2;
		B.Y:=Y2;
	end;
	Inc(SurfaceCount);
	{$IFDEF DEBUG}
	if SurfaceCount>MaxRect then
		MaxRect:=SurfaceCount;
	{$ENDIF}
end;

procedure TSurface.Remove(Pos: Integer);
begin
	{$IFDEF DEBUG}
	if (Pos<0) or (Pos>=SurfaceCount) then
		RunError(201);
	{$ENDIF}
	Dec(SurfaceCount);
	if Pos<SurfaceCount then
		Move(SurfaceRect^[Pos+1],SurfaceRect^[Pos],(SurfaceCount-Pos)*SizeOf(TRect));
	{$IFNDEF DPMI}
	if SurfaceCount=0 then
		Allocate(0)
	else
	{$ENDIF}
	if SurfaceCount<Allocated-2*UpSize then
		Allocate(Allocated-UpSize);
end;

constructor TSurface.Init(AWidth,AHeight: Integer; Visible: Boolean);
begin
	inherited Init;
	Width:=AWidth;
	Height:=AHeight;
	{$IFDEF CACHE}
	GetMem(Cache,Height*2);
	{$ENDIF}
	SurfaceClear(Visible);
	{$IFDEF DEBUG}
	Inc(UseCount);
	Inc(UseTotal);
	{$ENDIF}
end;

constructor TSurface.InitCopy(ASurface: PSurface);
begin
	Init(ASurface^.Width,ASurface^.Height,False);
	SurfaceCopy(ASurface);
end;

constructor TSurface.InitCopyInvert(ASurface: PSurface);
var
	I: Integer;
begin
	Init(ASurface^.Width,ASurface^.Height,True);
	for I:=0 to ASurface^.SurfaceCount-1 do
		with ASurface^.SurfaceRect^[I] do
			SurfaceBar(A.X,A.Y,B.X,B.Y,False);
end;

procedure TSurface.SurfaceCopy(ASurface: PSurface);
begin
	if ASurface<>@self then begin
		{$IFDEF CACHE}
		FreeMem(Cache,Height*2);
		{$ENDIF}
		SurfaceCount:=0;
		if ASurface^.SurfaceCount>0 then begin
			Allocate(ASurface^.Allocated);
			SurfaceCount:=ASurface^.SurfaceCount;
			Move(ASurface^.SurfaceRect^,SurfaceRect^,SurfaceCount*SizeOf(TRect));
		end else if Allocated>UpSize then
			Allocate(UpSize);
		Width:=ASurface^.Width;
		Height:=ASurface^.Height;
		{$IFDEF CACHE}
		GetMem(Cache,Height*2);
		InvalidCache:=ASurface^.InvalidCache;
		Move(ASurface^.Cache^,Cache^,2*InvalidCache);
		{$ENDIF}
	end;
end;

procedure TSurface.SurfaceSwap(ASurface: PSurface);
type
	Image=record
		Link: Word;
		SurfaceRect: record end;
	end;
var
	Buf: Pointer;
begin
	{$IFDEF CACHE}
	Caching:=True;
	{$ENDIF}
	Swap(Image(Pointer(@self)^).SurfaceRect,Image(Pointer(ASurface)^).SurfaceRect,SizeOf(TSurface)-SizeOf(Image));
	{$IFDEF CACHE}
	Caching:=False;
	{$ENDIF}
end;

procedure TSurface.SurfaceResize(AWidth,AHeight: Integer; Visible: Boolean);
var
	Temp: TSurface;
	OWidth,OHeight: Integer;
begin
	if (Width<>AWidth) or (Height<>AHeight) then begin
		if (AWidth<0) or (AHeight<0) then
			RunError;
		{$IFDEF CACHE}
		FreeMem(Cache,Height*2);
		GetMem(Cache,AHeight*2);
		InvCache(0);
		{$ENDIF}
		if (AWidth=0) or (AHeight=0) then begin
			Width:=AWidth;
			Height:=AHeight;
			SurfaceClear(False);
		end else if (Width=0) or (Height=0) then begin
			Width:=AWidth;
			Height:=AHeight;
			SurfaceClear(Visible);
		end else begin
			OWidth:=Width;
			OHeight:=Height;
			Width:=Maximum(AWidth,OWidth);
			Height:=Maximum(AHeight,OHeight);
			case (Ord(AWidth>OWidth)+Ord(AHeight>OHeight)*2) and (-Ord(Visible)) of
			0: begin {Visible=False or downsizing both}
					if AWidth<>OWidth then
						SurfaceBar(AWidth,0,OWidth,AHeight-1,False);
					if AHeight<>OHeight then
						SurfaceBar(0,AHeight,OWidth-1,OHeight-1,False);
				end;
			1: begin {Upsizing horz, downsizing vert}
					if AWidth<>OWidth then
						SurfaceBar(AWidth,0,OWidth,AHeight-1,True);
					if AHeight<>OHeight then
						SurfaceBar(0,AHeight,OWidth-1,OHeight-1,False);
				end;
			2: begin {Downsizing horz, upsizing vert}
					if AWidth<>OWidth then
						SurfaceBar(AWidth,0,OWidth,OHeight-1,False);
					if AHeight<>OHeight then
						SurfaceBar(0,AHeight,AWidth-1,OHeight-1,True);
				end;
			3: begin {Upsizing both}
					if AWidth<>OWidth then
						SurfaceBar(AWidth,0,OWidth,AHeight-1,True);
					if AHeight<>OHeight then
						SurfaceBar(0,AHeight,OWidth-1,OHeight-1,True);
				end;
			end;
			Width:=AWidth;
			Height:=AHeight;
		end;
	end;
end;

procedure TSurface.SurfaceCombine(ASurface: PSurface; X,Y: Integer; Mode: TCombineMode; Border: Boolean);
var
	I,Y2: Integer;
	Vis,UseTemp: Boolean;
	Temp: TSurface;
begin
{	if (not Assigned(ASurface) or (ASurface^.SurfaceCount=0) or (X+ASurface^.Width<=0) or (Y+ASurface^.Height<=0) or
			(X>=Width) or (Y>=Width)) and ((Mode<>cmXOR) or (not Border)) then begin
		if Border then
			SurfaceBorder(X,Y,Pred(X+ASurface^.Width),Pred(Y+ASurface^.Height),Mode in [cmOR,cmORNOT]);
		Exit;
	end;}
	if (Width=0) or (Height=0) then
		Exit;
	case Mode of
	cmOR,cmANDNOT: begin
			if (ASurface^.Width=0) or (ASurface^.Height=0) then begin
				if Border then
					SurfaceClear(Mode=cmOR);
				Exit;
			end;
			if (Mode=cmOR) and (SurfaceCount=1) then
				with SurfaceRect^[0] do
					if (A.X<=X) and (A.Y<=Y) and (B.X>X+ASurface^.Width) and (B.Y>Y+ASurface^.Height) then begin
						if Border then
							SurfaceClear(True);
						Exit;
					end;
			UseTemp:=ASurface=@self;
			if UseTemp then begin
				Temp.InitCopy(ASurface);
				ASurface:=@Temp;
			end;
			Vis:=Mode=cmOR;
			with ASurface^ do begin
				I:=0;
				Y2:=self.Height-Y;
				while I<SurfaceCount do begin
					with SurfaceRect^[I] do begin
						if A.Y>=Y2 then
							Break;
						self.SurfaceBar(A.X+X,A.Y+Y,B.X+X,B.Y+Y,Vis);
					end;
					Inc(I);
				end;
			end;
			if Border then
				SurfaceBorder(X,Y,Pred(X+ASurface^.Width),Pred(Y+ASurface^.Height),Vis);
			if UseTemp then
				Temp.Done;
		end;
	cmAND: if SurfaceCount>0 then begin
			Temp.InitCopyInvert(ASurface);
			SurfaceCombine(@Temp,X,Y,cmANDNOT,Border);
			Temp.Done;
		end;
	cmORNOT: begin
			Temp.InitCopyInvert(ASurface);
			SurfaceCombine(@Temp,X,Y,cmOR,Border);
			Temp.Done;
		end;
	cmXOR: begin
			Temp.InitCopy(@self);
			Temp.SurfaceCombine(ASurface,X,Y,cmAND,Border);
			SurfaceCombine(ASurface,X,Y,cmOR,Border);
			SurfaceCombine(@Temp,0,0,cmANDNOT,Border);
			Temp.Done;
		end;
	end;
end;

procedure TSurface.VisibleBar(Start: Integer; var ToCheck: Integer; X1,Y1,X2,Y2: Integer);
var
	I,X3,Y3,X4,Y4: Integer;
	Hor: Boolean;
begin
	I:=Start;
	while I<=ToCheck do begin
		with SurfaceRect^[I] do
			if Surfaces.Intersect(X1,Y1,X2,Y2,A.X,A.Y,B.X,B.Y,X3,Y3,X4,Y4) then begin
				if Y3>Y1 then
					VisibleBar(Succ(I),ToCheck,X1,Y1,X2,Y3-1);
				if Y4<Y2 then
					VisibleBar(Succ(I),ToCheck,X1,Y4+1,X2,Y2);
				Hor:=X3>X1;
				if Hor then
					VisibleBar(Succ(I),ToCheck,X1,Y3,X3-1,Y4);
				if X4<X2 then begin
					VisibleBar(Succ(I),ToCheck,X4+1,Y3,X2,Y4);
					Hor:=True;
				end;
				if Hor then begin
					if Y3>SurfaceRect^[I].A.Y then begin
						Inc(ToCheck);
						with SurfaceRect^[I] do {DO NOT REMOVE - SurfaceRect-POINTER MAY HAVE CHANGED!!!}
							Add(ToCheck,A.X,A.Y,B.X,Y3-1);
						SurfaceRect^[I].A.Y:=Y3;
					end;
					if Y4<SurfaceRect^[I].B.Y then begin
						Inc(ToCheck);
						with SurfaceRect^[I] do {DO NOT REMOVE - SurfaceRect-POINTER MAY HAVE CHANGED!!!}
							Add(ToCheck,A.X,Y4+1,B.X,B.Y);
						SurfaceRect^[I].B.Y:=Y4;
					end;
				end;
				Exit;
			end;
		Inc(I);
	end;
	Add(SurfaceCount,X1,Y1,X2,Y2);
end;

procedure TSurface.Sort(L,R: Integer);
var
	I,J: Integer;
	X: Longint;
begin
	I:=L;
	J:=R;
	X:=Longint(SurfaceRect^[(L+R) shr 1].A);
	repeat
		while Longint(SurfaceRect^[I].A)<X do
			Inc(I);
		while X<Longint(SurfaceRect^[J].A) do
			Dec(J);
		if I<=J then begin
			Swap(SurfaceRect^[I],SurfaceRect^[J],SizeOf(TRect));
			Inc(I);
			Dec(J);
		end;
	until I>J;
	if L<J then
		Sort(L,J);
	if I<R then
		Sort(I,R);
end;

procedure TSurface.SurfaceBar(X1,Y1,X2,Y2: Integer; Visible: Boolean);
var
	I,J,ToCheck,X3,Y3,X4,Y4: Integer;
	R: ^TRect;
begin
	if X1>X2 then begin
		I:=X1;
		X1:=X2;
		X2:=I;
	end;
	if Y1>Y2 then begin
		I:=Y1;
		Y1:=Y2;
		Y2:=I;
	end;
	if X1<0 then
		X1:=0;
	if Y1<0 then
		Y1:=0;
	if X2>=Width then
		X2:=Pred(Width);
	if Y2>=Height then
		Y2:=Pred(Height);
	if (X1>=Width) or (X2<0) or (Y1>=Height) or (Y2<0) then {Out of bounds}
		Exit;
	{$IFDEF CACHE}
	InvCache(Y1);
	{$ENDIF}
	if not Visible then begin  { +-------+ This is the works:}
		ToCheck:=SurfaceCount;   { +-+---+-+ we check for overcuts and then for slices as shown left.}
		I:=0;                    { | |   | | }
		while I<ToCheck do begin { +-+---+-+ }
			with SurfaceRect^[I] do{ +-------+ }
				if Surfaces.Intersect(X1,Y1,X2,Y2,A.X,A.Y,B.X,B.Y,X3,Y3,X4,Y4) then begin
					if Y3>A.Y then {cut top}
						Add(SurfaceCount,A.X,A.Y,B.X,Y3-1);
					with SurfaceRect^[I] do {DO NOT REMOVE - SurfaceRect-POINTER MAY HAVE CHANGED!!!}
						if Y4<B.Y then {cut bottom}
							Add(SurfaceCount,A.X,Y4+1,B.X,B.Y);
					with SurfaceRect^[I] do {DO NOT REMOVE - SurfaceRect-POINTER MAY HAVE CHANGED!!!}
						if X3>A.X then {cut left}
							Add(SurfaceCount,A.X,Y3,X3-1,Y4);
					with SurfaceRect^[I] do {DO NOT REMOVE - SurfaceRect-POINTER MAY HAVE CHANGED!!!}
						if X4<B.X then {cut right}
							Add(SurfaceCount,X4+1,Y3,B.X,Y4);
					Remove(I);
					Dec(ToCheck);
				end else if A.Y>Y2 then
					Break
				else
					Inc(I);
		end;
		if SurfaceCount=0 then {nothing there anymore?}
			Exit;
		Sort(0,Pred(SurfaceCount)); {now we must sort the list, Y first and X after}
	end else begin
		I:=0;
		while I<SurfaceCount do
			with SurfaceRect^[I] do
				if Surfaces.Intersect(X1,Y1,X2,Y2,A.X,A.Y,B.X,B.Y,X3,Y3,X4,Y4) and
						(A.X=X3) and (A.Y=Y3) and (B.X=X4) and (B.Y=Y4) then
					Remove(I)
				else
					Inc(I);
		ToCheck:=Pred(SurfaceCount);
		VisibleBar(0,ToCheck,X1,Y1,X2,Y2);
		if SurfaceCount=0 then {nothing there anymore?}
			Exit;
		Sort(0,Pred(SurfaceCount)); {now we must sort the list, Y first and X after}
		I:=0; {split if necessary so that we have proper stripes}
		while (I<SurfaceCount-1) do begin
			if (Succ(SurfaceRect^[I].B.X)=SurfaceRect^[I+1].A.X) and (SurfaceRect^[I].A.Y=SurfaceRect^[I+1].A.Y) then begin
				Y1:=SurfaceRect^[I].B.Y-SurfaceRect^[I+1].B.Y;
				if Y1<>0 then begin
					if Y1<0 then begin
						Y1:=SurfaceRect^[I].B.Y+1;
						Inc(I);
						ToCheck:=I+1;
					end else begin
						ToCheck:=I-1;
						Y1:=SurfaceRect^[I+1].B.Y+1;
					end;
					J:=I+1;
					while (J<SurfaceCount) and (Y1>SurfaceRect^[J].A.Y) do
						Inc(J);
					while (J<SurfaceCount) and (Y1=SurfaceRect^[J].A.Y) and (SurfaceRect^[I].A.X>=SurfaceRect^[J].A.X) do
						Inc(J);
					with SurfaceRect^[I] do
						Add(J,A.X,Y1,B.X,B.Y);
					SurfaceRect^[I].B.Y:=Pred(Y1);
					I:=ToCheck;
				end else
					Inc(I);
			end else
				Inc(I);
		end;
		J:=0; {now the X-fragments have to be collected and glued}
		repeat
			I:=Succ(J);
			while (I<SurfaceCount) do begin
				R:=@SurfaceRect^[I];
				with SurfaceRect^[J] do
					if (R^.A.Y=A.Y) and (R^.B.Y=B.Y) and (Succ(B.X)=R^.A.X) then begin
						B.X:=R^.B.X;
						Remove(I);
					end else
						Break;
			end;
			Inc(J);
		until J>=SurfaceCount;
	end;
	J:=0; {now the Y-fragments have to be collected and glued}
	repeat
		I:=Succ(J);
		repeat
			while (I<SurfaceCount) and (SurfaceRect^[J].B.Y>=SurfaceRect^[I].A.Y) do
				Inc(I);
			if I>=SurfaceCount then
				Break;
			R:=@SurfaceRect^[I];
			with SurfaceRect^[J] do begin {collate in Y-direction}
				if Succ(B.Y)<>R^.A.Y then
					Break;
				if (R^.A.X=A.X) and (R^.B.X=B.X) then begin
					B.Y:=R^.B.Y;
					Remove(I);
				end else
					Inc(I);
			end;
		until False;
		Inc(J);
	until J>=SurfaceCount;
end;

procedure TSurface.SurfaceBorder(X1,Y1,X2,Y2: Integer; Visible: Boolean);
var {This proc will be optimized some day...}
	I: Integer;
begin
	if X1>X2 then begin
		I:=X1;
		X1:=X2;
		X2:=I;
	end;
	if Y1>Y2 then begin
		I:=Y1;
		Y1:=Y2;
		Y2:=I;
	end;
	SurfaceBar(-1,-1,Width,Y1-1,Visible);
	SurfaceBar(-1,Y1,X1-1,Y2,Visible);
	SurfaceBar(X2+1,Y1,Width,Y2,Visible);
	SurfaceBar(-1,Y2+1,Width,Height,Visible);
end;

procedure TSurface.SurfaceInvert;
var
	Temp: TSurface;
begin
	Temp.InitCopyInvert(@self);
	SurfaceSwap(@Temp);
	Temp.Done;
end;

procedure TSurface.SurfaceClear(Visible: Boolean);
begin
	{$IFDEF CACHE}
	InvCache(0);
	{$ENDIF}
	if SurfaceCount>=1 then begin
		SurfaceCount:=1;
		Remove(0);
	end;
	if Visible and (Width>0) and (Height>0) then
		Add(0,0,0,Pred(Width),Pred(Height));
end;

function TSurface.SurfaceEqual(ASurface: PSurface): Boolean;
type
	PIntArray=^TIntArray;
	TIntArray=array[0..32766] of Integer;
var
	I: Integer;
begin
	if SurfaceCount=ASurface^.SurfaceCount then begin
		for I:=0 to Pred(SurfaceCount*4) do
			if PIntArray(SurfaceRect)^[I]<>PIntArray(ASurface^.SurfaceRect)^[I] then begin
				SurfaceEqual:=False;
				Exit;
			end;
		SurfaceEqual:=True;
	end else
		SurfaceEqual:=False;
end;

procedure TSurface.SurfaceResetHLine(var AHLine: THLine);
var
	I,Y,X1,X2{$IFDEF CACHE},Nr{$ENDIF}: Integer;
begin
	{$IFDEF CACHE}
	Nr:=GetCache(AHLine.A.Y);
	if Nr=SurfaceCount then begin
		AHLine.B.Y:=Nr;
		Exit;
	end;
	{$ENDIF}
	with AHLine do begin
		if A.X>B.X then begin
			I:=A.X;
			A.X:=B.X;
			B.X:=I;
		end;
		if (A.X>=Width) or (B.X<0) then begin
			B.Y:=SurfaceCount;
			Exit;
		end;
		if A.X<0 then
			A.X:=0;
		if B.X>=Width then
			B.X:=Pred(Width);
		X1:=A.X;
		Y:=A.Y;
		X2:=B.X;
		{$IFDEF CACHE}
		I:=Nr;
		{$ELSE}
		I:=0;
		{$ENDIF}
		while I<SurfaceCount do
			with SurfaceRect^[I] do
				if A.Y>Y then begin
					AHLine.B.Y:=SurfaceCount;
					Exit;
				end else if (Y>B.Y) or (X2<A.X) or (X1>B.X) then
					Inc(I)
				else
					Break;
		B.Y:=I;
	end;
end;

function TSurface.SurfaceGetHLine(var AHLine: THLine; var X1,X2: Integer): Boolean;
var
	HX1,Y,HX2,I: Integer;
begin
	with AHLine do begin
		HX1:=A.X;
		Y:=A.Y;
		HX2:=B.X;
		I:=B.Y;
	end;
	if I>=SurfaceCount then begin
		SurfaceGetHLine:=False;
		Exit;
	end;
	with SurfaceRect^[I] do begin
		X1:=Maximum(HX1,A.X);
		X2:=Minimum(HX2,B.X);
	end;
	Inc(I);
	while I<SurfaceCount do
		with SurfaceRect^[I] do begin
			if A.Y>Y then begin
				I:=SurfaceCount;
				Break;
			end;
			if (Y>B.Y) or (HX2<A.X) or (HX1>B.X) then
				Inc(I)
			else
				Break;
		end;
	AHLine.B.Y:=I;
	SurfaceGetHLine:=True;
end;

function TSurface.RectOversect(X1,Y1,X2,Y2: Integer): Boolean;
var
	I: Integer;
begin
	if X1>X2 then begin
		I:=X1;
		X1:=X2;
		X2:=I;
	end;
	if Y1>Y2 then begin
		I:=Y1;
		Y1:=Y2;
		Y2:=I;
	end;
	if (X2>=0) and (Y2>=0) and (X1<Width) and (Y1<Height) then begin
		{$IFDEF CACHE}
		I:=GetCache(Y1);
		if I=SurfaceCount then
		{$ENDIF}
		I:=0;
		while I<SurfaceCount do begin
			with SurfaceRect^[I] do
				if A.Y>Y2 then
					Break
				else if (X1<=B.X) and (X2>=A.X) and (Y1<=B.Y) then begin
					RectOversect:=True;
					Exit;
				end;
			Inc(I);
		end;
	end;
	RectOversect:=False;
end;

function TSurface.RectIncluded(X1,Y1,X2,Y2: Integer): Boolean;
var
	I: Integer;
begin
	if X1>X2 then begin
		I:=X1;
		X1:=X2;
		X2:=I;
	end;
	if Y1>Y2 then begin
		I:=Y1;
		Y1:=Y2;
		Y2:=I;
	end;
	if (X2>=0) and (Y2>=0) and (X1<Width) and (Y1<Height) then begin
		{$IFDEF CACHE}
		I:=GetCache(Y1);
		if I=SurfaceCount then
		{$ENDIF}
		I:=0;
		while I<SurfaceCount do begin
			with SurfaceRect^[I] do
				if (Y2<=B.Y) and (X1>=A.X) and (X1<=B.X) then begin
					if X2>B.X then
						Break;
					if Y2<=B.Y then begin
						RectIncluded:=True;
						Exit;
					end;
				end else if	A.Y>Y2 then
					Break;
			Inc(I);
		end;
	end;
	RectIncluded:=False;
end;

{function TSurface.PointIncluded(X,Y: Integer): Boolean; assembler;
asm
	les di,self
	mov cx,es:[di].SurfaceCount
	jcxz @False
	mov ax,X
	cmp ax,0
	jl @False
	cmp ax,es:[di].Width
	jge @False
	mov bx,Y
	cmp bx,es:[di].Height
	jge @False
	les di,es:[di].SurfaceRect
	jmp @StartLoop
@FindRect:
	dec cx
	jz @False
	add di,type TRect
@StartLoop:
	cmp es:[di].TRect.B.Y,bx
	jl @FindRect
@Check:
	cmp es:[di].TRect.A.Y,bx
	jg @False
	cmp es:[di].TRect.A.X,ax
	jg @FindRect
	cmp es:[di].TRect.B.X,ax
	jl @FindRect
@EndCheck:
	mov ax,1
	jmp @End
@False:
	sub ax,ax
@End:
end;}

function TSurface.PointIncluded(X,Y: Integer): Boolean;
var
	I: Integer;
begin
	if (X>=0) and (Y>=0) and (X<Width) and (Y<Height) then begin
		{$IFDEF CACHE}
		I:=GetCache(Y);
		{$ELSE}
		I:=0;
		{$ENDIF}
		while I<SurfaceCount do begin
			with SurfaceRect^[I] do
				if A.Y>Y then
					Break
				else if (X>=A.X) and (X<=B.X) and (Y<=B.Y) then begin
					PointIncluded:=True;
					Exit;
				end;
			Inc(I);
		end;
	end;
	PointIncluded:=False;
end;

function TSurface.SurfaceRectHLineCut(Y: Integer): Integer;
var
	I,J,First: Integer;
begin
	if (Y<0) or (Y>=Height) then
		Exit; {Out of range!}
	{$IFDEF CACHE}
	InvCache(0);
	{$ENDIF}
	I:=0;
	First:=SurfaceCount;
	while (I<SurfaceCount) and (SurfaceRect^[I].A.Y<Y) do begin
		if (SurfaceRect^[I].B.Y>=Y) then begin
			J:=Succ(I);
			while (J<SurfaceCount) and ((Y>SurfaceRect^[J].A.Y) or
						((Y=SurfaceRect^[J].A.Y) and (SurfaceRect^[I].A.X>SurfaceRect^[J].B.X))) do
				Inc(J);
			with SurfaceRect^[I] do
				Add(J,A.X,Y,B.X,B.Y);
			SurfaceRect^[I].B.Y:=Y-1;
			if J<=First then
				First:=J-1;
		end;
		Inc(I);
	end;
	SurfaceRectHLineCut:=First;
	Sort(0,Pred(SurfaceCount));
end;

destructor TSurface.Done;
begin
	{$IFDEF DEBUG}
	Dec(UseCount);
	{$ENDIF}
	SurfaceCount:=0;
	Allocate(0);
	{$IFDEF CACHE}
	FreeMem(Cache,Height*2);
	{$ENDIF}
	inherited Done;
end;

end.